import pytest
import responses

from django.contrib.auth import authenticate
from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory
from mixer.backend.django import mixer

from . import utils as test_utils
from .. import exceptions
from .. import settings
from .. import utils

pytestmark = pytest.mark.django_db

BASE_URL = settings.API_BASE_URL
ACCESS_TOKEN_URL = f'{BASE_URL}/oauth/access_token'
DEBUG_TOKEN_URL = f'{BASE_URL}/debug_token'
ME_URL = f'{BASE_URL}/me'


def test_success_handler_default():
    req = RequestFactory().get('/')
    req.user = AnonymousUser()
    SessionMiddleware().process_request(req)
    req.session.save()

    # before our test, the request user is anonymous
    assert req.user.is_authenticated is False

    user = mixer.blend('auth.User')
    user.backend = 'facebook_login.auth_backends.FacebookAuthBackend'
    res = utils.success_handler_default(req, user)
    # after our test, the request has the authenticated user attached
    assert req.user == user
    assert req.user.is_authenticated is True


@responses.activate
def test_get_app_access_token():
    test_utils.add_response(ACCESS_TOKEN_URL, {'access_token': 'asd'})
    res = utils.get_app_access_token()
    assert res == 'asd', (
        "Should return a dict of the form `{'access_token': 'asd'}`")


@responses.activate
def test_get_app_access_token_error():
    test_utils.add_response(ACCESS_TOKEN_URL, {'access_token': ''})
    with pytest.raises(exceptions.AccessTokenException):
        # Should raise exception if no access_token could be retrieved
        utils.get_app_access_token()


@responses.activate
def test_get_app_access_token_error2():
    # no responses setup, so responses will return a ConnectionError
    with pytest.raises(exceptions.AccessTokenException):
        # Should raise exception if no access_token could be retrieved
        utils.get_app_access_token()


@responses.activate
def test_debug_token():
    test_utils.add_response(DEBUG_TOKEN_URL, {'data': {'user_id': '123'}})
    res = utils.debug_token('valid-token', 'valid-token')
    assert res['data']['user_id'] == '123', (
        "Should return a dict of the form `{'data': {'user_id': '123'}}`")


@responses.activate
def test_debug_token_error1():
    test_utils.add_response(
        DEBUG_TOKEN_URL, {'error': {
            'message': 'Error message from Facebook'
        }})
    try:
        res = utils.debug_token('valid-token', 'valid-token')
    except exceptions.FacebookRequestException as ex:
        assert str(ex) == 'Error message from Facebook', (
            'If Facebook returned an error message, it should be set in this'
            ' exception')


@responses.activate
def test_debug_token_error2():
    # no responses setup, so responses will return a ConnectionError
    with pytest.raises(exceptions.FacebookNetworkException):
        res = utils.debug_token('valid-token', 'valid-token')


@responses.activate
def test_debug_token_error3():
    test_utils.add_response(
        DEBUG_TOKEN_URL,
        {'data': {
            'error': {
                'message': 'Error message from Facebook'
            }
        }})
    with pytest.raises(exceptions.DebugTokenException):
        res = utils.debug_token('valid-token', 'valid-token')


@responses.activate
def test_get_user_email():
    test_utils.add_response(ME_URL, {'email': 'test@example.com'})
    res = utils.get_user_email('valid-token')
    assert res == 'test@example.com', ('Should return the Facebook user email')


@responses.activate
def test_get_user_email_error1():
    test_utils.add_response(ME_URL, {'id': '123'})
    with pytest.raises(exceptions.UserEmailException):
        res = utils.get_user_email('valid-token')


@responses.activate
def test_get_user_email_error2():
    # no responses setup, so responses will return a ConnectionError
    with pytest.raises(exceptions.FacebookNetworkException):
        res = utils.get_user_email('valid-token')


@responses.activate
def test_get_user_email_error3():
    test_utils.add_response(ME_URL, {'error': 'Error message'})
    with pytest.raises(exceptions.FacebookRequestException):
        res = utils.get_user_email('bad-token')



import pytest
import responses

from django.contrib.auth.models import User
from mixer.backend.django import mixer

from . import utils as test_utils
from .. import settings
from .. import auth_backends
from .. import models
from .. import exceptions

pytestmark = pytest.mark.django_db

URLS = test_utils.URLS


class TestFacebookAuthBackend:

    def test_authenticate_none(self):
        backend = auth_backends.FacebookAuthBackend()
        req = test_utils.get_request()

        res = backend.authenticate(req, token=None)
        assert res is None, ('Should return `None`, if no token is given')

    @responses.activate
    def test_authenticate_bad_token(self):
        test_utils.add_response(URLS['access'], {'access_token': 'asd'})
        test_utils.add_response(URLS['debug'],
                                {'error': {
                                    'message': 'Bad Token'
                                }})
        backend = auth_backends.FacebookAuthBackend()
        req = test_utils.get_request()

        with pytest.raises(exceptions.FacebookRequestException):
            backend.authenticate(req, token='test-user-token')

    @responses.activate
    def test_authenticate_no_email_permission(self):
        test_utils.add_response(URLS['access'], {'access_token': 'asd'})
        test_utils.add_response(URLS['debug'], {'data': {'user_id': '123'}})
        test_utils.add_response(URLS['me'], {'id': '123'})
        backend = auth_backends.FacebookAuthBackend()
        req = test_utils.get_request()

        with pytest.raises(exceptions.UserEmailException):
            result = backend.authenticate(req, token='test-user-token')

    @responses.activate
    def test_authenticate_known_facebook_user(self):
        fb_user_id = '123'
        test_utils.add_response(URLS['access'], {'access_token': 'asd'})
        test_utils.add_response(URLS['debug'],
                                {'data': {
                                    'user_id': fb_user_id
                                }})
        test_utils.add_response(URLS['me'], {'email': 'test@example.com'})
        backend = auth_backends.FacebookAuthBackend()
        req = test_utils.get_request()

        user = mixer.blend('auth.User', email='test@example.com')
        fb_account = mixer.blend(
            'facebook_login.FacebookAccount', user=user, fb_user_id=fb_user_id)

        result = backend.authenticate(req, token='test-user-token')
        assert result == user, (
            'Should return the Django user that has the same email as the'
            ' given Facebook user')
        assert models.FacebookAccount.objects.all().count() == 1, (
            'Should create a FacebookAccount instance')
        assert models.FacebookAccount.objects.all()[0].user == user, (
            'Should tie FacebookAccount instance to Django user')

    @responses.activate
    def test_authenticate_connect_fb_user_with_known_django_user(self):
        test_utils.add_response(URLS['access'], {'access_token': 'asd'})
        test_utils.add_response(URLS['debug'], {'data': {'user_id': '123'}})
        test_utils.add_response(URLS['me'], {'email': 'test@example.com'})
        backend = auth_backends.FacebookAuthBackend()
        req = test_utils.get_request()

        user = mixer.blend('auth.User', email='test@example.com')
        result = backend.authenticate(req, token='test-user-token')
        assert result == user, ('Should return the Django user ')

    @responses.activate
    def test_authenticate_signup_completely_unknown_user(self):
        test_utils.add_response(URLS['access'], {'access_token': 'asd'})
        test_utils.add_response(URLS['debug'], {'data': {'user_id': '123'}})
        test_utils.add_response(URLS['me'], {'email': 'test@example.com'})
        backend = auth_backends.FacebookAuthBackend()
        req = test_utils.get_request()

        result = backend.authenticate(req, token='test-user-token')
        assert User.objects.all().count() == 1, (
            'Should create new Django user')
        assert models.FacebookAccount.objects.all().count() == 1, (
            'Should create new FacebookAccount')
        assert models.FacebookAccount.objects.all()[
            0].user == User.objects.all()[0], (
                'Should tie new instances together')

    def test_get_user(self):
        backend = auth_backends.FacebookAuthBackend()
        user = mixer.blend('auth.User')
        res = backend.get_user(666)
        assert res is None, (
            'Should return `None` when an unknown user-id was provided')

        res = backend.get_user(user.pk)
        assert res == user, ('Should return the user for the given PK')



# Generated by Django 2.0.8 on 2018-09-20 03:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('facebook_login', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='facebookaccount',
            name='fb_user_id',
            field=models.CharField(max_length=1024, unique=True),
        ),
    ]



from builtins import str
from builtins import object
import graphene
import json
import requests

from django.contrib.auth import authenticate, get_user_model

from .models import FacebookAccount
from . import utils


class FacebookAuthMutation(graphene.Mutation):

    class Arguments(object):
        access_token = graphene.String()

    status = graphene.Int()
    form_errors = graphene.String()
    extra = graphene.String()

    @staticmethod
    def mutate(root, info, **args):
        user_access_token = args.get('access_token')

        if user_access_token is None:
            return FacebookAuthMutation(
                status=400,
                form_errors=json.dumps({
                    'facebook': ['No user access token was provided']
                }),
                extra=None,
            )

        try:
            user = authenticate(request=info.context, token=user_access_token)
        except Exception as ex:
            return utils.error_handler(FacebookAuthMutation, info.context, ex)

        extra = utils.success_handler(info.context, user)

        return FacebookAuthMutation(status=200, form_errors=None, extra=extra)


class Mutation(object):
    facebook_auth = FacebookAuthMutation.Field()



from django.utils.translation import ugettext as _


class AccessTokenException(Exception):
    """
    Is raised when we are unable to retrieve an access token.

    This can happen either because Facebook is down or because our
    local_settings.py has wrong values for APP_ID and APP_SECRET.

    This is a severe error and means that our engineers need to fix the
    configuration.

    """

    def __str__(self):
        return _("Error retrieving access token. Please try again later.")


class DebugTokenException(Exception):
    """
    Is raised when we do get a response from Facebook but the response json
    has an `error` key.

    This would probably mean that someone is trying to attack the system by
    submitting invalid Facebook user access tokens.

    This should never happen and can probably be ignored.

    """

    def __init__(self, error):
        self.error = error

    def __str__(self):
        try:
            return self.error['message']
        except KeyError:
            return _('Unable to validate token. Please try again later.')


class UserEmailException(Exception):
    """
    Is raised when the call to `/me` does not return a value for the email.

    This would probably mean that the user disabled the email-permission.

    This could also mean that the user does not have a primary email in their
    Facebook profile, but we think this is such a rare case, this library
    can't currently handle that.

    """

    def __str__(self):
        return _(
            "Email access was not granted. Please try to login again and grant"
            " email access.")


class FacebookRequestException(Exception):
    """
    Is raised when requests return a 400 response status.

    This usually means that you provided a wrong app access token.

    This is a severe error and means that our engineers need to make sure that
    the app is properly configured.

    """

    def __init__(self, error):
        self.error = error

    def __str__(self):
        try:
            return self.error['message']
        except KeyError:
            return _("Unknown error occurred when communicating with Facebook.")


class FacebookNetworkException(Exception):
    """
    Is raised when requests returns any exception while making a request to the
    Facebook API.

    This would mean that the Facebook API is down or that our server has
    connection problems.

    This is a severe error and means that our engineers need to make sure that
    the issue is not on our side.

    """

    def __str__(self):
        return _(
            "Unable to communicate with Facebook. Please try again later.")



import requests
import json

from django.contrib.auth import authenticate, login
from django.utils.module_loading import import_string

from . import settings
from . import exceptions


def error_handler_default(facebook_auth_mutation, request, exception):
    """
    Returns the mutation with form_errors.

    """
    return facebook_auth_mutation(
        status=400,
        form_errors=json.dumps({
            'facebook': [str(exception)]
        }),
        extra=None,
    )


def error_handler(facebook_auth_mutation, request,
                  exception):  # pragma: nocover
    # too trivial to test
    """Imports and then calls the error handler function."""
    error_handler_func = import_string(settings.ERROR_HANDLER)
    return error_handler_func(facebook_auth_mutation, request, exception)


def success_handler_default(request, user):
    """
    Logs in the user that has been queried or created by the
    FacebookAuthMutation.

    """
    login(request, user)
    return None


def success_handler(request, user):  # pragma: nocover
    # too trivial to test
    """Imports and then calls the success handler function."""
    success_handler_func = import_string(settings.SUCCESS_HANDLER)
    return success_handler_func(request, user)


def get_app_access_token():
    """
    Gets an access token from Facebook.

    We need this token in order to make more Facebook API requests.

    """
    try:
        resp = requests.get(f'{settings.API_BASE_URL}/oauth/access_token?'
                            f'client_id={settings.APP_ID}&'
                            f'client_secret={settings.APP_SECRET}&'
                            f'grant_type=client_credentials').json()
    except Exception:
        resp = {}
    if not 'access_token' in resp or not resp['access_token']:
        raise exceptions.AccessTokenException()
    return resp['access_token']


def debug_token(user_access_token, app_access_token):
    """
    Validates a given user access token and returns user information.

    On the client, the Facebook login returns a userID and a token. We only
    need the userID to hook it up with our own user model but we need to verify
    that the token that the client has sent us is indeed valid, otherwise
    anyone could impersonate anyone by just sending us their Facebook userID.

    """
    try:
        resp = requests.get(f'{settings.API_BASE_URL}/debug_token?'
                            f'input_token={user_access_token}&'
                            f'access_token={app_access_token}')
    except Exception as ex:
        raise exceptions.FacebookNetworkException()
    resp = resp.json()
    if 'error' in resp:
        raise exceptions.FacebookRequestException(resp['error'])
    if 'error' in resp['data']:
        raise exceptions.DebugTokenException(resp['data']['error'])
    return resp


def get_user_email(user_access_token):
    """
    Given a valid user access token, it will return the user's email.

    """
    try:
        resp = requests.get(f'{settings.API_BASE_URL}/me?'
                            f'access_token={user_access_token}&fields=email')
    except Exception:
        raise exceptions.FacebookNetworkException()
    resp = resp.json()
    if 'error' in resp:
        raise exceptions.FacebookRequestException(resp['error'])
    email = resp.get('email', None)
    if email in [None, '']:
        raise exceptions.UserEmailException()
    return email



import pytest

from .. import exceptions


class TestAccessTokenException:

    def test_exception(self):
        ex = exceptions.AccessTokenException()
        assert 'Error retrieving' in str(ex)


class TestDebugTokenException:

    def test_exception(self):
        ex = exceptions.DebugTokenException({})
        assert 'Unable to' in str(ex)

        ex = exceptions.DebugTokenException({'message': 'Foo'})
        assert str(ex) == 'Foo'


class TestUserEmailException:

    def test_exception(self):
        ex = exceptions.UserEmailException()
        assert 'Email access' in str(ex)


class TestFacebookRequestException:

    def test_exception(self):
        ex = exceptions.FacebookRequestException({})
        assert 'Unknown error' in str(ex)

        ex = exceptions.FacebookRequestException({'message': 'Foo'})
        assert str(ex) == 'Foo'


class TestFacebookNetworkException:

    def test_exception(self):
        ex = exceptions.FacebookNetworkException()
        assert 'Unable to' in str(ex)
